use proc_macro::{ Delimiter, Group, Punct, Ident, Spacing, Span, TokenStream, TokenTree };

#[macro_use] mod memo;
#[macro_use] mod quote;

memoize! {
    fn generics_and_tuple(count: usize) -> (TokenStream, TokenStream) {
        if count != 1 {
            let mut stream = TokenStream::new();

            (0..)
                .filter_map(|number| char::from_u32(number))
                .filter(|character| character.is_alphabetic())
                .filter(|character| character.is_uppercase())
                .map(|character| format!("{character}"))
                .take(count)
                .for_each(|ident| {
                    stream.extend([
                        TokenTree::from(Ident::new(&ident, Span::mixed_site())),
                        TokenTree::from(Punct::new(',', Spacing::Alone)),
                    ].into_iter().collect::<TokenStream>());
                });

            (
                stream.clone(),
                TokenStream::from(TokenTree::from(Group::new(Delimiter::Parenthesis, stream))),
            )
        } else {
            (
                quote! { T, },
                quote! { (T,) },
            )
        }
    }
}

fn impl_sealed(count: usize) -> TokenStream {
    let (generics, tuple) = generics_and_tuple(count);
    
    quote! {
        impl<(@generics)> Sealed for (@tuple) { }
    }
}

fn impl_foo(count: usize) -> TokenStream {
    let (generics, tuple) = generics_and_tuple(count);
    
    quote! {
        impl<(@generics)> Foo for (@tuple) { }
    }
}

#[proc_macro]
pub fn example(_: TokenStream) -> TokenStream {
    let mut output = TokenStream::new();

    for i in 0..=12 {
        output.extend(impl_sealed(i));
        output.extend(impl_foo(i));
    }

    output
}
